(load "util.lisp")
(load "input.lisp")
(load "czasownik.lisp")
(load "przyimek.lisp")
(load "rzeczownik.lisp")
(load "przymiotnik.lisp")
(load "przyslowek.lisp")
(load "spojnik.lisp")
(load "zaimek.lisp")
(load-dictionaries)

;; Czci mowy

(defvar pos-table nil)
(setq pos-table
  '(("czasownik"  get-verb analyze-and-learn-verb)
;    ("czstka"   get-particle)
;    ("liczebnik" get-number)
    ("przyimek"    get-preposition)
    ("przymiotnik"  get-adjective analyze-and-learn-adjective)
    ("przyswek" get-adverb analyze-and-learn-adverb)       
    ("rzeczownik" get-substantive analyze-and-learn-substantive)
    ("spjnik"    get-conjunction)
    ("zaimek"     get-pronoun)))

(defun slowo (word)
  (mapcan
   #'(lambda (drv)
       (let ((tag (intern (car drv))))
	 (mapcan
	  #'(lambda (x)
	      (when x
		(list
		 (if (listp x)
		     (cons tag x)
		   (list tag x)))))
	  (funcall (cadr drv) word))))
   pos-table))

(defun learn-word (word)
  (format t "** Analiza \"~A\"~%" word)
  (let ((variants))
    (dolist (drv pos-table)
      (let ((variant (funcall (cadr drv) word)))
	(when variant
	  (format t "~A~%" (car drv))
	  (print-result variant)
	  (setq variants (cons (car drv) variants)))))

    (when (or (not variants) (getyn "Jeszcze warianty?"))
      (let ((n 1))
	(dolist (drv pos-table)
	  (format t "(~A)~A " n (car drv))
	  (setq n (1+ n)))
	(let ((drv (nth (1- (getnumber ": Cz mowy?" 1 (length pos-table)))
			    pos-table)))
	  (when (caddr drv)
	    (funcall (caddr drv) word))))))
  
  (format t "**~%"))

(defun learn-from-text (file)
  (dolist (s (with-open-file (str file :direction :input)
			     (tokenize-stream str)))
    (dolist (word (car s))
      (learn-word word))))

(defun lex-text (file)
  (mapcar
   #'(lambda (sent)
       (mapcar
	#'(lambda (x)
	    (or (slowo x) (list (list '|niewiadome| x))))
	(car sent)))
   (with-open-file (str file :direction :input)
		   (tokenize-stream str))))

(defmacro update-concord (obj val)
  "Dodaje do list uzgodnie obiektu OBJ warto VAL"
  `(setf (get 'concord ,obj) (cons ,val (get 'concord ,obj))))

(defun concord-subst-with-prep (sentence cases pos)
  "Szuka rzeczownika ktryby si uzgadnia z przyimkiem w pozycji POS,
ktry wymaga przypadkw CASES"
  (do ((tail sentence (cdr tail))
       (i 0 (1+ i)))
      ((not tail))
    (let ((it (car tail)))
      ;; Szukamy rzeczownika
      (case (car it)
	(|rzeczownik|
	 (let ((cs (intersection
		    (to-list (subst-get-case (cdr it)))
		    (to-list cases))))
	   ;; Jeli ten zgadza si z wymaganymi przypadkami, uaktualniamy list
	   (when cs
	     ;; Mamy niepusty wynik, zamieszczamy go
	     (let ((tmp (copy-list (cdr it))))
	       (subst-set-case tmp cs)
	       (let ((r (cons '|rzeczownik| tmp)))
		 (update-concord r pos)
		 (rplaca tail r)
		 (return i))))))
	((|spjnik| |przyswek| |zaimek| |niewiadome| |przymiotnik|))
	((otherwise)
	 (return))))))

(defun remove-part-from-sentence-recursive (lsent n upd-list &optional removed-list)
  "Funkcja pomocnicza dla remove-part-from-sentence.
Usuwa ze zdania (CAR LSENT) element N oraz wszystkie uzgodnione z nim elementy.
Uaktualnia odpowiednio listy CONCORD pozostaych elementw oraz numery w licie
UPD-LIST tak, by wci zawieray one prawidowe pozycji sw w reszcie zdania.

Zamieszcza reszt zdania w (CAR LSENT). Zwraca list usunitych elementw.
Opcjonalny argument REMOVED-LIST uywa si do przekazania tej listy pomidzy
wywoaniami rekurencyjnymi.
"
  (let ((sent (car lsent))) ;; Zdanie z ktrym mamy do czynienia

    ;; 1: Uaktualnienie list 'CONCORD elementw zdania. Jeli jaka pozycja
    ;;    w licie jest wiksza ni N, naley j zmniejszy o jednostk.
    (dolist (elt sent)
      (let ((concord (mapcan
		      #'(lambda (cn)
			  (cond
			   ((> cn n)
			    (list (1- cn)))
			   ((= cn n)
			    nil)
			   (t
			    (list cn))))
		      (get 'concord elt))))
	(setf (get 'concord elt) concord)))

    ;; 2. Uaktualnienie numerw w UPD-LIST na tej samej zasadzie
    (do ((tail upd-list (cdr tail)))
	((not tail))
      (when (> (car tail) n)
	(rplaca tail (1- (car tail)))))

    ;; 3. Wyuszczenie elementu N i uzgodnionych z nim elementw
    (format t "REMOVING ~A from ~A~%" n sent);HERE 
    (let* ((removed-elt (nth n sent))
	   (removed-list (cons removed-elt removed-list)))
      (format t "REMOVED-ELT ~A ~A~%" removed-elt (get 'concord removed-elt));HERE 
      (setf (car lsent) (remove-nth sent (list n))) ; Uaktualniamy zdanie
      (dolist (c (get 'concord removed-elt))  ;; Iteracja w licie 'CONCORD
	(when (> c n)
	  (setq c (1- c)))
	(setq removed-list
	      (remove-part-from-sentence-recursive lsent c upd-list
						   removed-list)))
      (setf (get 'concord removed-elt) nil)  ; Uzgodnienia ju nie maj sensu
      removed-list)))

(defun remove-part-from-sentence (sent n &key update)
"Usuwa ze zdania SENT element N oraz wszystkie uzgodnione z nim elementy.
Uaktualnia odpowiednio listy CONCORD pozostaych elementw oraz numery w
opcjonalnej licie UPDATE tak, by wci zawieray one prawidowe pozycji
sw w reszcie zdania.

Zwraca CONS ktrego CAR zawiera list usunitych elementw, za CDR - reszt
zdania.
"
  (let* ((lsent (list sent))
	 (removed-list (remove-part-from-sentence-recursive lsent n update)))
    (cons removed-list
	  (car lsent))))

(defun concord-adj-with-subst (subst sentence pos &key count)
  "Szuka przymiotnika ktryby si zgadza z rzeczownikiem SUBST, znajdujcym
si w pozycji POS w zdaniu SENTENCE"
; (format t "SENT ~A ~A~%" sentence pos)
  (do ((cases (subst-get-case (cdr subst)))
       (nums (subst-get-number (cdr subst)))
       (tail sentence (cdr tail))
       (i 0 (1+ i)))
      ((or (not tail) (and count (= i count))))
;    (format t "IT ~A~%" (car tail))
    (let ((it (car tail)))
      (case (car it)
	((|przymiotnik|)
	 (format t "Przymiotnik ~A~%" it)
	 (let ((cs (intersection
		    (to-list (adj-get-case (cdr it)))
		    (to-list cases)))
	       (ns (intersection
		    (to-list (adj-get-number (cdr it)))
		    (to-list nums))))
	   ;; Jeli ten zgadza si z wymaganymi przypadkami, uaktualniamy list
	   (when (and cs ns)
	     ;; Mamy niepusty wynik, zamieszczamy go
	     (subst-set-case (cdr subst) cs)
	     (subst-set-number (cdr subst) ns)
	     (adj-set-case (cdr it) cs)
	     (adj-set-number (cdr it) ns)
	     (update-concord it pos)
;	     (format t "ITEM ~A C ~A~%" it (get 'concord it))
	     (return i))))

	((|spjnik|)
	 (format t "Spjnik ~A~%" it)
	 (unless (or (string-equal (cadr it) "i")
		     (string-equal (cadr it) "oraz")
		     (string-equal (cadr it) "a")
		     (string-equal (cadr it) "take")
		     (string-equal (cadr it) "zarwno")
		     (string-equal (cadr it) "czy")
		     (string-equal (cadr it) "albo")
		     (string-equal (cadr it) "bd")
		     (string-equal (cadr it) "lub"))
	   (return)))

	((otherwise) ;; FIXME co z tym?
;	 (format t "RET~%")
	 (return))))))

(defun concord-subst-with-pron (sentence pron pos)
  (let ((number (if (eq (car (cdr pron)) 'ZD)
		    (nth 4 (cdr pron))
		  (pronoun-get-number (cdr pron))))
	(case (pronoun-get-case (cdr pron))))
    (when (and case number)
      (do ((tail sentence (cdr tail))
	   (i 0 (1+ i)))
	  ((not tail))
	(let ((it (car tail)))
	  (case (car it)
	    (|rzeczownik|
	     (let ((cs (intersection
			(to-list (subst-get-case (cdr it)))
			(to-list case)))
		   (ns (intersection
			(to-list (subst-get-number (cdr it)))
			(to-list number))))
	       (when (and cs ns)
		 (pronoun-set-case (cdr pron) cs)
		 (pronoun-set-number (cdr pron) ns)
		 (subst-set-case (cdr it) cs)
		 (subst-set-number (cdr it) ns)
		 (update-concord it pos)
		 (return i))))))))))

(defun solve-concordance (sentence)
  "Ustala uzgodnienia w zdaniu SENTENCE. Zdanie ma by w formie normalnej,
tzn. kady jego element jest list (cz-mowy definicja). List otrzyman
na wyjciu lex-text mona rozwin za dopomog funkcji (make-variants)"
  (do ((tail sentence (cdr tail))
       (i 0 (1+ i)))
      ((not tail) sentence)
    (let ((it (car tail)))
      (case (car it)
	((|spjnik| |przyswek|) ) ; Ignorowane
	(|przyimek|
	 (let ((res (concord-subst-with-prep (cdr tail)
					     (prep-get-case (cdr it)) i)))
	   (if res
	       (update-concord it (+ i res 1))
	     (return))))
	(|zaimek|
	 (let ((res (concord-subst-with-pron (cdr tail) it i)))
	   (if res
	       (update-concord it (+ i res 1)))))

	(|rzeczownik|
	 (let ((res (concord-adj-with-subst it (cdr tail) i)))
	   (when res
	     (update-concord it (+ i res 1))))
         (do* ((next 0 (+ 1 res next))
	       (res (concord-adj-with-subst it sentence i :count i)
		    (when (and (< next i)
			       (eq (car (nth next sentence)) '|przymiotnik|))
		      (concord-adj-with-subst it
					      (nthcdr next sentence) i
					      :count (- i next)))))
	     ((not res))
	   (update-concord it (+ res next))))
	((otherwise)
	 ) ; Na razie ignorowane
	))))

(defun count-concordances (sent)
  "Zwraca liczb uzgodnie w zdaniu SENT"
  (do ((count 0)
       (lst sent (cdr lst)))
      ((not lst) count)
    (when (get 'concord (car lst))
      (setq count (1+ count)))))

(defun check-concordances (sentence &key weight max)
  "Sprawdza rne warianty uzgodnie w zdaniu SENTENCE. Zwraca list
moliwych wariantw. Warianty listy s odsortowane wedug iloci uzgodnie:
im wicej sow uzgadnia si midzy sob, tym bliej do pocztku listy znajduje
si ten wariant.

Opcionalny argument :max ustala maksymaln ilo zwrconych wariantw.

Opcionalny argument :weight ustala minimaln \"wag\" uzgodnie.
"
  (let ((res (sort
	      (mapcan
	       #'(lambda (sent)
		   (when (solve-concordance sent)
		     (list
		      (cons (count-concordances sent)
			    sent))))
	       (make-variants sentence))
	      #'(lambda (a b)
		  (> (car a) (car b))))))
    (cond
     (weight
      (mapcan
       #'(lambda (x)
	   (when (>= (car x) weight)
	     (cdr x)))
       res))
     ((and max (< max (length res)))
      (mapcar #'cdr (butlast res (- (length res) max))))
     (t
      (mapcar #'cdr res)))))

(defun subject-p (part)
  "Zwraca T jeli PART (opis sowa w zdaniu) *moe* by podmiotem"
  (or
   (and (eq (car part) '|rzeczownik|)
	(member 'M (to-list (subst-get-case (cdr part)))))
   (and (eq (car part) '|zaimek|)
	(not (get 'concord part))
	(subjective-pronoun-p (cdr part)))))

(defun find-subject-candidates (sent)
  "Szuka kandydatw na podmiot w zdaniu SENT. Zwraca list
pozycji w zdaniu"
  (do ((n 0 (1+ n))
       (ret)
       (tail sent (cdr tail)))
      ((not tail) ret)
    (when (subject-p (car tail))
      (push n ret))))

(defun word-get-number (word)
  "Zwraca liczb (kategoria gramatyczna), do ktrej naley sowo WORD.
Jeli sowo nie posiada kategorii liczby, zwraca NIL"
  (case (car word)
    (|rzeczownik| (subst-get-number (cdr word)))
    (|zaimek| (pronoun-get-number (cdr word)))))

;; get-subject-verb
(defun get-sv (sent)
  "Wyuszcza grupy podmiota i orzeczenia. Zwraca list:

   (N-PODMIOT N-ORZECZENIE)

N - pozycja w zdaniu"
  (let ((candidates (find-subject-candidates sent)))
    (if candidates
	(mapcan
	 #'(lambda (subject-n)
	     (let ((x (do ((num (word-get-number (nth subject-n sent)))
			   (tail sent (cdr tail))
			   (i 0 (1+ i)))
			  ((not tail))
			(when (and (not (= i subject-n))
				   (eq (caar tail) '|czasownik|)
				   (or (not num)
				       (intersection (to-list num)
					(to-list (verb-get-number (cdar tail))))))
			  ;; Czasownik pasuje
			  (return i)))))
	       (when x
		 (list (cons subject-n x)))))
	 candidates)
      ;; Podmiot domylny
      (do ((tail sent (cdr tail))
	   (i 0 (1+ i))
	   (ret))
	  ((not tail) ret)
	(when (and (eq (caar tail) '|czasownik|) ;; Szukamy czasownika
		   (verb-get-tense (cdar tail))) ;; ... w formie osobowej
	  (push (cons nil i) ret))))))

(defun derive-subject (verb)
  "Wywodzi podmiot domylny z orzeczenia VERB"
  (get-pronoun
   (case (verb-get-person verb)
     ((1) (if (eq (verb-get-number verb) 'LP) "ja" "my"))
     ((2) (if (eq (verb-get-number verb) 'LP) "ty" "wy"))
     ((3) (if (eq (verb-get-number verb) 'LP) "on" "oni")))))

(defun get-predicate (sent vp)
  "Rozbija zdanie na podmiot, i grup orzeczenia. Zwraca list

   (PODMIOT GRUPA-ORZECZENIA RESZTA)

FIXME: Orzeczenie imienne"
  (if (= vp (- (length sent) 1))
      (cons (nth vp sent)
	    (remove-nth sent (list vp)))
    (let ((next (nth (1+ vp) sent)))
      (cond
       ((and (eq (car next) '|czasownik|)
	     (not (verb-get-tense (cdr next)))) ;; bezokolicznik
	(setf (get 'concord (nth vp sent)) (list (1+ vp))
	      (get 'concord (nth (1+ vp) sent)) (list vp))
	(remove-part-from-sentence sent vp))
       (t
	(remove-part-from-sentence sent vp))))))

(defun get-object (sent)
  "Szuka dopenienia bliszego w zdaniu SENT. Zwraca CONS ktrego CAR
zawiera znalezione dopenienie (lub NIL, jeli go nie ma), a CDR - reszt
zdania."
  (do ((tail sent (cdr tail))
       (i 0 (1+ i)))
      ((not tail) (cons nil sent))
    ;; Dopenienie blisze - to rzeczownik bd zaimek w bierniku (i
    ;; wszystkie uzgodnione z nim sowa)
    (when (or
	   (and (eq (caar tail) '|rzeczownik|)
		(member 'B (to-list (subst-get-case (cdar tail)))))
	   (and (eq (caar tail) '|zaimek|)
		(member 'B (to-list (pronoun-get-case (cdar tail))))))
      ;; Przed zwrceniem wartoci ustawiamy pola przypadku wszystkich
      ;; elementw dopenienia w 'B (biernik)
;;      (format t "FOUND ~A ~A~%" i sent)
      (let ((ret (remove-part-from-sentence sent i)))
	(dolist (elt (car ret))
	  (case (car elt)
	    (|rzeczownik|  (subst-set-case (cdr elt) 'B))
	    (|przymiotnik| (adj-set-case (cdr elt) 'B))
	    (|zaimek|      (pronoun-set-case (cdr elt) 'B))))
;;	(format t "FO ~A~%" ret)
	(return ret)))))

(defun czy-p (word)
  (and (eq (car word) '|spjnik|) ;; FIXME: partykua
       (string= (cadr word) "czy")))

(defun normalize-struct (str)
  "Normalizuje struktur zdania, wyuszczajc ewentualn cz
orzeczenia imiennego i umieszczajc j w polu dopenienia.
Orzeczeniem imiennym nazywa si konstrukcja
    by + rzeczownik w mianowniku bd narzdniku
Np.: To _jest kot_
     On _jest czowiekiem_
"  
  (cond
   ((and
     (eq (car (nth 1 str)) '|czasownik|)
     (string= (cadr (nth 1 str)) "by")
     (> (length (nth 0 str)) 1)
     (eq (caar (nth 0 str)) '|zaimek|)
     (eq (pronoun-get-type (cdar (nth 0 str))) 'ZW)
     (member 'RN (to-list (pronoun-get-gender (cdar (nth 0 str))))))
    (let ((x (cdr (nth 0 str))))
      (rplacd (nth 0 str) nil)
      (setf (nth 2 str) x))))
  str)

(defun copy-sentence (sent)
  "Zwraca dokadn kopi zdania SENT"
  (map 'list
       #'(lambda (orig copy)
	   (setf (get 'concord copy) (get 'concord orig))
	   copy)
       sent
       (copy-tree sent)))

(defun analyze-sentence-struct (sent)
  "Analizuje struktur zdania SENT. Zwraca list (patrz get-predicate).
Waciwo TYPE tej listy zawiera typ zdania: Oglne lub szczegowe"
  (let ((type (cond
	       ((czy-p (car sent))
		(setq sent (cdr sent))
		'OGLNE)
	       (t
		'SZCZEGӣOWE))))
    (mapcan
     #'(lambda (pair)
	 (let* ((sp (car pair))
		(vp (cdr pair))
		(sent (copy-sentence sent))
		(subj (if sp
			  (let* ((upd (list vp))
				 (tmp (remove-part-from-sentence sent sp :update upd)))
			    (setq sent (cdr tmp)
				  vp (car upd))
			    (car tmp))
			(derive-subject (cdr (nth vp sent))))))
	   (let ((pred (get-predicate sent vp)))
	     (list
	      (let ((struct (cons
			     subj
			     (cons
			      (car pred)
			      (get-object (cdr pred))))))
		(setf (get 'type struct) type)
		(normalize-struct struct))))))
     (get-sv sent))))

(defun analyze-sentence-from-text (file)
  "Analiza *pierwszego* zdania z pliku FILE.
Polega na wywoaniu (analyze-sentence-struct) dla kadego zdania
zwrconego przez (check-concordances).

Wynik sortujemy wg. istnienia/nieistnienia dopenienia oraz liczby
uzgodnie w dopenieniu.
"
  (sort
   (mapcan
    #'analyze-sentence-struct
    (check-concordances (car (lex-text file)) :max 8)) 
   #'(lambda (x y)
       (cond
	((and (nth 2 x) (not (nth 2 y))) t)
	((and (not (nth 2 x)) (nth 2 y)) nil)
	(t (< (count-concordances (nth 2 x)) (count-concordances (nth 2 y))))))))

	 