;; Lista nazw przypadkw
(defvar case-table
  '((M . "Mianownik")
    (D . "Dopeniacz")
    (C . "Celownik")
    (B . "Biernik")
    (W . "Woacz")
    (N . "Narzdnik")
    (Mc . "Miejscownik")))

;;; Akcesory
;; Rzeczownik
(defmacro subst-get-root (s)
  `(nth 0 ,s))

(defmacro subst-get-dict (s)
  `(nth 1 ,s))

(defmacro subst-get-decl (s)
  `(nth 2 ,s))

(defmacro subst-get-case (s)
  `(nth 3 ,s))

(defmacro subst-get-number (s)
  `(nth 4 ,s))

(defmacro subst-set-case (s val)
  `(progn (setf (subst-get-case ,s) ,val) ,s))

(defmacro subst-set-number (s val)
  `(progn (setf (subst-get-number ,s) ,val) ,s))

;; Czasownik
(defmacro verb-get-root (s)
  `(nth 0 ,s))

(defmacro verb-get-conj (s)
  `(nth 1 ,s))

(defmacro verb-get-tense (s)
  `(nth 2 ,s))

(defmacro verb-get-person (s)
  `(nth 3 ,s))

(defmacro verb-get-number (s)
  `(nth 4 ,s))

;; Przyimek
(defmacro prep-get-form (s)
  `(car ,s))

(defmacro prep-get-case (s)
  `(cdr ,s))

;; Przymiotnik
;;  '(forma-sownikowa rodzaj przypadek liczba)
(defmacro adj-get-dictform (entry)
  `(nth 0 ,entry))

(defmacro adj-get-gender (entry)
  `(nth 1 ,entry))

(defmacro adj-get-case (entry)
  `(nth 2 ,entry))

(defmacro adj-get-number (entry)
  `(nth 3 ,entry))

(defmacro adj-set-case (s val)
  `(progn (setf (adj-get-case ,s) ,val) ,s))

(defmacro adj-set-number (s val)
  `(progn (setf (adj-get-number ,s) ,val) ,s))

;; Zaimek (p. zaimek.lisp)
(defmacro pronoun-get-type (p)
  `(nth 0 ,p))
(defmacro pronoun-has-person (p)
  `(member (pronoun-get-type ,p) '(ZO ZZ ZD)))
(defmacro pronoun-get-person (p)
  `(and (pronoun-has-person ,p) (nth 1 ,p)))
(defmacro pronoun-get-case (p)
  `(nth (if (pronoun-has-person ,p)
	    2
	  1) ,p))
(defmacro pronoun-get-number (p)
  `(case (pronoun-get-type ,p)
     ((ZO ZD) (nth 3 ,p))
     ((ZP ZPR ZW ZOK) (nth 2 ,p))))

(defmacro pronoun-get-gender (p)
  `(case (pronoun-get-type ,p)
     ((ZO) (nth 4 ,p))
     ((ZP ZW) (nth 3 ,p))))

(defmacro pronoun-set-case (s val)
  `(progn (setf (pronoun-get-case ,s) ,val) ,s))

(defmacro pronoun-set-number (s val)
  `(case (pronoun-get-type ,s)
     ((ZO ZD) (progn (setf (nth 3 ,s) ,val) ,s))
     ((ZP ZPR ZW ZOK)  (progn (setf (nth 2 ,s) ,val) ,s))))

;;;
(defmacro while (test &body body)
  `(do ()
       ((not ,test) t)
       ,@body))

;;
(defun to-list (d)
  (if (listp d)
      d
    (list d)))

(defun remove-nth (seq nlist)
  "Usuwa z listy SEQ elementy o numerach z listy NLIST. Zwraca
modyfikowan list"
  (do ((ret)
       (i 0 (1+ i))
       (tail seq (cdr tail)))
      ((not tail) (nreverse ret))
    (unless (member i nlist)
      (push (car tail) ret))))

;; Rozwija list DIGEST wedug jej elementu NUM.
;; FIXME do czasownik.lisp: (split-on-conj vd) == (burst-digest vd 1)
(defun burst-digest (digest num)
  (let ((x (nth num digest)))
    (if (listp x)
	(let ((head (butlast digest (- (length digest) num)))
	      (tail (nthcdr (1+ num) digest)))
	  (mapcan
	   #'(lambda (c)
	       (list
		(append head (list c) tail)))
	   x))
      (list digest))))

(defun make-variants (data)
  (when data
    (let* ((cell (car data))
	   (tail (cdr data)))
      (mapcan
       #'(lambda (x)
	   (if tail
	       (mapcar
		#'(lambda (y)
		    (cons (copy-list x) (copy-list y)))
		(make-variants tail))
	     (list (list (copy-list x)))))
       (copy-list cell)))))

(defun print-result (r)
  (format t "~{~A ~}~%" r))

(defun match-ending (word pattern &optional ignore-len)
  "Zwraca T jeeli kocwk sowa WORD jest PATTERN.
Jeeli podano opcjonalny parametr IGNORE-LEN, ignoruje podan ilo
liter od koca sowa.

   (match-ending \"imionami\" \"ami\") => T
   (match-ending \"imionami\" \"on\" 3) => T"
  (let ((wlen (- (length word) (or ignore-len 0)))
	(plen (length pattern)))
    (and
     (>= wlen plen)
     (string-equal (substring word (- wlen plen) wlen) pattern))))

(defun remove-ending (word ending-length)
  "Usuwa ostatnie ENDING-END liter ze sowa WORD"
  (substring word 0 (- (length word) ending-length)))

(defun assqf (key lst fun)
  "To samo co ASSOC, z t rnic e uywa funkcji FUN do porwna"
  (dolist (elt lst)
	  (when (funcall fun key (car elt))
		(return elt))))

(defun vocal-ending (word)
  "Zwraca T jeli WORD koczy si na samogosk."
  (let ((len (length word)))
    (member (aref word (1- len))
	    '(#\a #\ #\e #\ #\i #\o #\ #\u #\y))))

(defun consonant-ending (word)
  "Zwraca T jeli WORD koczy si na spgosk."
  (not (vocal-ending word)))

;; FIXME
(defun hard-ending (word)
  "Zwraca T jeli WORD ma kocwk tward"
  (let ((len (length word)))
    (cond
     ((and (> (length word) 2) (match-ending word "i"))
      (not (member (aref word (- len 2))
		   '(#\m #\n #\s #\c #\z))))
     ((> (length word) 1)
      (member (aref word (1- len))
	      '(#\b #\d #\f #\g #\ #\m #\n #\p #\r #\s #\t #\w #\z))))))

(defun soft-ending (word)
  "Zwraca T jeli WORD ma kocwk mikk"
  (not (hard-ending word)))

(defun analyze-word (word drivers)
  "Zwraca list moliwych form sowa WORD wedug listy kierujcej DRIVERS.
Poprawne formy naley wybiera ze sownika"
  (dolist (drv drivers)
    (when (match-ending word (car drv))
      (let* ((ending-len (length (car drv)))
	     (root (remove-ending word ending-len)))
	(let ((r  (mapcan #'(lambda (x)
			      (let ((r (funcall x word root ending-len)))
				(when r
				  (list r))))
			  (cdr drv))))
	  (when r
	    (return r)))))))


;; Dialog

(defun prompted-read-line (prompt)
  (format t "~A " prompt)
  (read-line nil nil))

(defun get-list-answer (prompt)
  (do ((ret (prompted-read-line prompt) (prompted-read-line prompt)))
      ((or (null ret) (= (length ret) 0)) ret)
    (cond
     ((char= (aref ret 0) #\()
      (if (not (char= (aref ret (1- (length ret))) #\)))
	  (format t "Niezakoczona lista. Sprobuj jeszcze raz~%")
	(return (read-from-string ret))))
     ((char= (aref ret (1- (length ret))) #\))
      (format t "Niepoprawna skadnia.~%"))
     (t
      (return (read-from-string (string-concat "(" ret ")")))))))

(defun getyn (prompt)
  (do ((ret (prompted-read-line prompt) (prompted-read-line prompt)))
      (nil)
    (if (> (length ret) 0)
	(let ((c (aref ret 0)))
	  (cond
	   ((or (char= c #\t) (char= c #\T))
	    (return t))
	   ((or (char= c #\n) (char= c #\N))
	    (return nil)))))
    (format t "Prosz odpowiedzie \"Tak\" (T) czy \"Nie\" (N)~%")))

(defun getstring (prompt)
  (prompted-read-line prompt))

(defun getsymbol (prompt)
  (let ((s (getstring prompt)))
    (and s
	 (intern (string-upcase s)))))

(defun getnumber (prompt &optional min max)
  (do ((ret (prompted-read-line prompt) (prompted-read-line prompt)))
      ((null ret) ret)
    (let ((val (read-from-string ret)))
      (if (numberp val)
	(cond
	 ((and min (< val min))
	  (format t "Niepoprawna liczba (< ~A).~%" min))
	 ((and max (> val max))
	  (format t "Niepoprawna liczba (> ~A).~%" max))
	 (t
	  (return val)))
	(format t "Prosz poda liczb.~%")))))
  


;; ----------------------------------------------------------------------------
;; Inicjalizacja
;;

(defun make-rc-name (&optional name)
  (if name name
    (namestring (merge-pathnames
		 (user-homedir-pathname)
		 (pathname ".slowo.rc")))))

(defun load-dictionaries (&optional name)
  (let* ((stream (open (make-rc-name name) :direction :input
		       :if-does-not-exist nil)))
    (cond
     (stream
      (do
       ((form (read stream nil nil) (read stream nil nil)))
       ((not form) nil)
       (eval form))
      (close stream)
      t)
     (t
      nil))))

(defun save-dictionaries (&optional name)
  (if *save-vars*
      (let (file)
	(setq file (open (make-rc-name name)
			 :direction :output
			 :if-does-not-exist :create :if-exists :supersede))
	(dolist (var *save-vars*)
		(format file "~&(setq ~D '" var)
		(write (symbol-value var) :closure : :stream file :pretty : )
		(format file ")~%"))
	(close file))))


(defmacro dict-store (dict entry)
  `(setq ,dict (cons ,entry ,dict)))



(defvar *save-vars* (list
		     'subst-dictionary
		     'subst-exceptions
		     'verb-present
		     'adjective-dictionary
		     'adverb-dictionary))

